const std = @import("std");

pub const std_options = std.Options{
    .log_level = .info,
};

pub fn main() !void {
    var gpa = std.heap.GeneralPurposeAllocator(.{}){};
    defer _ = gpa.deinit();
    const allocator = gpa.allocator();

    var circuit = Circuit.init(allocator);
    defer circuit.deinit();

    var microchip = try circuit.addComponent(Microchip);

    var not1 = try circuit.addComponent(Not);
    var not2 = try microchip.circuit.addComponent(Not);
    not1.invert_output = false;
    try microchip.resizeInputs(1);
    try microchip.resizeOutputs(1);
    try microchip.connectInput(0, &not2.component, 0);
    try microchip.connectOutput(&not2.component, 0, 0);

    try circuit.tick();
    std.log.info("{}", .{microchip.component.getOutput(0)});
    try circuit.tick();
    std.log.info("{}", .{microchip.component.getOutput(0)});
    try circuit.tick();
    std.log.info("{}", .{microchip.component.getOutput(0)});
}

test "basic circuit" {
    var input = Signal{ .digital = 1, .analog = 0.5 };

    var circuit = Circuit.init(std.testing.allocator);
    defer circuit.deinit();

    var not1 = try circuit.addComponent(Not);
    not1.invert_output = false;
    not1.component.inputs.items[0].signal = &input; // manually set the input here

    var battery1 = try circuit.addComponent(Battery);
    var battery2 = try circuit.addComponent(Battery);
    var battery3 = try circuit.addComponent(Battery);
    battery1.value = -0.5;
    battery2.value = 0.5;
    battery3.value = -1.0;

    var or1 = try circuit.addComponent(Or);
    var or2 = try circuit.addComponent(Or);
    or1.arithmetic_mode = true;
    or2.arithmetic_mode = true;
    try or1.component.resizeInputs(std.testing.allocator, 3);
    try or2.component.resizeInputs(std.testing.allocator, 3);

    var or3 = try circuit.addComponent(Or);
    var or4 = try circuit.addComponent(Or);
    var or5 = try circuit.addComponent(Or);

    var or6 = try circuit.addComponent(Or);
    var or7 = try circuit.addComponent(Or);
    var or8 = try circuit.addComponent(Or);
    var and1 = try circuit.addComponent(And);
    var and2 = try circuit.addComponent(And);
    or6.arithmetic_mode = true;
    or7.arithmetic_mode = true;
    or8.arithmetic_mode = true;
    and1.arithmetic_mode = true;
    and2.arithmetic_mode = true;

    try circuit.connectComponents(&battery1.component, 0, &or3.component, 1);
    try circuit.connectComponents(&battery1.component, 0, &or4.component, 1);
    try circuit.connectComponents(&battery1.component, 0, &or5.component, 1);

    try circuit.connectComponents(&battery2.component, 0, &or6.component, 0);
    try circuit.connectComponents(&battery2.component, 0, &or7.component, 0);
    try circuit.connectComponents(&battery2.component, 0, &or8.component, 0);

    try circuit.connectComponents(&battery3.component, 0, &and1.component, 0);
    try circuit.connectComponents(&battery3.component, 0, &and2.component, 0);

    try circuit.connectComponents(&not1.component, 0, &or3.component, 0);
    try circuit.connectComponents(&not1.component, 0, &or1.component, 0);
    try circuit.connectComponents(&not1.component, 0, &or1.component, 1);

    try circuit.connectComponents(&or1.component, 0, &or4.component, 0);
    try circuit.connectComponents(&or1.component, 0, &or2.component, 0);
    try circuit.connectComponents(&or1.component, 0, &or2.component, 1);
    try circuit.connectComponents(&or2.component, 0, &or5.component, 0);

    try circuit.connectComponents(&or3.component, 0, &or6.component, 1);
    try circuit.connectComponents(&or4.component, 0, &or7.component, 1);
    try circuit.connectComponents(&or5.component, 0, &or8.component, 1);

    try circuit.connectComponents(&or6.component, 0, &and1.component, 1);
    try circuit.connectComponents(&and1.component, 0, &or1.component, 2);
    try circuit.connectComponents(&or7.component, 0, &and2.component, 1);
    try circuit.connectComponents(&and2.component, 0, &or2.component, 2);

    try circuit.tick();

    try std.testing.expectEqual(Signal{ .analog = 0.0 }, or1.component.getOutput(0).*);
    try std.testing.expectEqual(Signal{ .analog = 0.0 }, or2.component.getOutput(0).*);

    try std.testing.expectEqual(Signal{ .analog = 0.5 }, or3.component.getOutput(0).*);
    try std.testing.expectEqual(Signal{ .analog = -0.5 }, or4.component.getOutput(0).*);
    try std.testing.expectEqual(Signal{ .analog = -0.5 }, or5.component.getOutput(0).*);

    try std.testing.expectEqual(Signal{ .analog = 1.0 }, or6.component.getOutput(0).*);
    try std.testing.expectEqual(Signal{ .analog = -1.0 }, and1.component.getOutput(0).*);
    try std.testing.expectEqual(Signal{ .analog = 0.0 }, or7.component.getOutput(0).*);
    try std.testing.expectEqual(Signal{ .analog = -0.0 }, and2.component.getOutput(0).*);
    try std.testing.expectEqual(Signal{ .analog = 0.0 }, or8.component.getOutput(0).*);
}

pub const Circuit = struct {
    allocator: std.mem.Allocator,
    components: Components,

    update_order: ?UpdateOrder = null,

    const Components = std.ArrayListUnmanaged(*Component);
    const ComponentIndex = u32;

    pub fn init(allocator: std.mem.Allocator) Circuit {
        return .{
            .allocator = allocator,
            .components = Components.empty,
        };
    }

    pub fn deinit(self: *Circuit) void {
        if (self.update_order) |p| self.allocator.free(p);
        for (0..self.components.items.len) |i| self.components.items[i].deinit(self.allocator);
        self.components.deinit(self.allocator);
    }

    pub fn addComponent(self: *Circuit, comptime T: type) !*T {
        var c = try T.init(self.allocator);
        errdefer c.component.deinit(self.allocator);
        if (self.update_order) |p| self.allocator.free(p);
        try self.components.append(self.allocator, &c.component);
        return c;
    }

    pub fn connectComponents(self: *Circuit, from: *Component, from_idx: usize, to: *Component, to_idx: usize) !void {
        try from.connect(self.allocator, from_idx, to, to_idx);
        if (self.update_order) |p| self.allocator.free(p);
    }

    pub fn tick(self: *Circuit) !void {
        if (self.update_order == null) try self.calculateUpdateOrder();
        for (self.update_order.?) |component| {
            std.log.debug("updating {}@{d}", .{ component, self.componentIndex(component).? });
            try component.update();
        }
    }

    fn calculateUpdateOrder(self: *Circuit) !void {
        var update_order = std.ArrayList(*Component).init(self.allocator);
        errdefer update_order.deinit();

        const visited = try self.allocator.alloc(bool, self.components.items.len);
        defer self.allocator.free(visited);

        // find components with nonexistant or disconnected inputs
        var source_components = std.ArrayList(ComponentIndex).init(self.allocator);
        defer source_components.deinit();
        source: for (self.components.items) |component| {
            for (component.inputs.items) |input|
                if (input.connection != null)
                    continue :source;

            const idx = self.componentIndex(component).?;
            try update_order.append(component);
            visited[idx] = true;
            try source_components.append(idx);
        }

        // handle the edge case of pure loops
        // TODO this is a HACKY temporary solution!
        if (source_components.items.len == 0 and self.components.items.len > 0) {
            try update_order.append(self.components.items[0]);
            try source_components.append(0);
            visited[0] = true;
        }

        // do multiple Depth First Searches (with extra steps) once per source component
        for (source_components.items) |source_component_idx| {
            var stack = std.ArrayList(ComponentIndex).init(self.allocator);
            defer stack.deinit();

            var component_idx = source_component_idx;

            dfs: while (true) {
                // check each output of the current component
                for (self.components.items[component_idx].outputs.items) |output| {
                    // check each connection of the current output
                    connections: for (output.connections.items) |connection| {
                        const new_idx = self.componentIndex(connection.component).?;

                        // check if component is "ready"
                        // i.e. it hasn't already been added to the process list, but all its inputs have.
                        if (visited[new_idx]) continue :connections;
                        for (connection.component.inputs.items) |input| {
                            if (input.connection) |input_connection| {
                                if (!visited[self.componentIndex(input_connection.component).?]) {
                                    continue :connections;
                                }
                            }
                        }

                        // component is ready, continue the DFS down that branch
                        try update_order.append(connection.component);
                        try stack.append(new_idx);
                        visited[new_idx] = true;
                        component_idx = new_idx;
                        continue :dfs;
                    }
                }

                component_idx = stack.pop() orelse break :dfs;
            }
        }

        if (self.update_order) |p| self.allocator.free(p);
        self.update_order = try update_order.toOwnedSlice();
    }

    fn componentIndex(self: *Circuit, component: *Component) ?ComponentIndex {
        const idx = std.mem.indexOfScalar(*Component, self.components.items, component);
        if (idx) |i|
            return @intCast(i)
        else
            return null;
    }

    const UpdateOrder = []*Component;
};

pub const Component = struct {
    name: []const u8,

    inputs: Inputs,
    outputs: Outputs,

    updateFn: *const fn (*Component) AllocatorError!void,
    deinitFn: *const fn (*Component, std.mem.Allocator) void,

    pub const Input = struct {
        signal: ?*Signal = null,
        connection: ?Connection = null,
    };
    pub const Output = struct {
        signal: Signal = .{},
        connections: Connections = .empty,
    };

    pub const Connection = struct {
        component: *Component,
        idx: usize,
    };
    // TODO wouldn't this need to be a list of ?Connection since an
    pub const Connections = std.ArrayListUnmanaged(Connection);
    const Inputs = std.ArrayListUnmanaged(Input);
    const Outputs = std.ArrayListUnmanaged(Output);

    pub fn init(
        self: *Component,
        allocator: std.mem.Allocator,
        name: []const u8,
        inputs_len: usize,
        outputs_len: usize,
        updateFn: *const fn (*Component) AllocatorError!void,
        deinitFn: *const fn (*Component, std.mem.Allocator) void,
    ) !void {
        var inputs = Inputs.empty;
        errdefer inputs.deinit(allocator);
        try inputs.resize(allocator, inputs_len);
        for (0..inputs.items.len) |i| inputs.items[i] = .{};

        var outputs = Outputs.empty;
        errdefer outputs.deinit(allocator);
        try outputs.resize(allocator, outputs_len);
        for (0..outputs.items.len) |i| outputs.items[i] = .{};

        self.name = name;
        self.inputs = inputs;
        self.outputs = outputs;
        self.updateFn = updateFn;
        self.deinitFn = deinitFn;
    }

    pub fn deinit(self: *Component, allocator: std.mem.Allocator) void {
        self.inputs.deinit(allocator);
        for (self.outputs.items) |*output|
            output.connections.deinit(allocator);
        self.outputs.deinit(allocator);
        self.deinitFn(self, allocator);
    }

    pub inline fn update(self: *Component) AllocatorError!void {
        try self.updateFn(self);
    }

    // TODO allow inserting the new elements at an arbitrary index
    // TODO ensure this won't break the opposide side's connections
    pub fn resizeInputs(self: *Component, allocator: std.mem.Allocator, new_len: usize) !void {
        const old_len = self.inputs.items.len;
        try self.inputs.resize(allocator, new_len);
        if (new_len > old_len) for (old_len..new_len) |i| {
            self.inputs.items[i] = .{};
        };
    }

    // TODO allow inserting the new elements at an arbitrary index
    // TODO ensure this won't break the opposide side's connections
    pub fn resizeOutputs(self: *Component, allocator: std.mem.Allocator, new_len: usize) !void {
        const old_len = self.outputs.items.len;
        try self.outputs.resize(allocator, new_len);
        if (new_len > old_len) for (old_len..new_len) |i| {
            self.outputs.items[i] = .{};
        };
    }

    pub fn getInput(self: *Component, idx: usize) ?*Signal {
        return self.inputs.items[idx].signal;
    }

    pub fn getOutput(self: *Component, idx: usize) *Signal {
        return &self.outputs.items[idx].signal;
    }

    fn connect(self: *Component, allocator: std.mem.Allocator, self_idx: usize, to: *Component, to_idx: usize) !void {
        const input = &to.inputs.items[to_idx];
        input.signal = self.getOutput(self_idx);
        input.connection = .{
            .component = self,
            .idx = self_idx,
        };

        const output = &self.outputs.items[self_idx];
        try output.connections.append(allocator, .{
            .component = to,
            .idx = to_idx,
        });
    }

    pub fn format(
        self: *Component,
        comptime _: []const u8,
        _: std.fmt.FormatOptions,
        writer: anytype,
    ) !void {
        try writer.print("Component{{ .name = \"{s}\", .inputs.items.len = {d}, .outputs.items.len = {d} }}", .{
            self.name,
            self.inputs.items.len,
            self.outputs.items.len,
        });
    }
};

pub const Signal = struct {
    digital: i2 = 0,
    analog: f32 = 0.0,
    color: u24 = 0,

    pub fn format(
        self: Signal,
        comptime fmt: []const u8,
        options: std.fmt.FormatOptions,
        writer: anytype,
    ) !void {
        _ = .{ fmt, options };
        try writer.writeAll("Signal(");
        if (self.digital < 0) try writer.writeByte('-') else try writer.writeByte('+');
        try writer.print("{d} / {d:0>1.4})", .{
            @abs(self.digital),
            self.analog,
        });
    }
};

pub const Microchip = struct {
    component: Component,

    circuit: Circuit,
    outer_to_inner: std.ArrayListUnmanaged(?Component.Connection),
    inner_to_outer: std.ArrayListUnmanaged(?Component.Connection),

    pub fn init(allocator: std.mem.Allocator) !*Microchip {
        var self = try allocator.create(Microchip);
        errdefer allocator.destroy(self);
        self.circuit = Circuit.init(allocator);
        self.outer_to_inner = .empty;
        self.inner_to_outer = .empty;
        errdefer self.circuit.deinit();
        try Component.init(&self.component, allocator, "Microchip", 0, 0, &update, &deinit);
        return self;
    }

    pub fn deinit(component: *Component, allocator: std.mem.Allocator) void {
        const self: *Microchip = @fieldParentPtr("component", component);
        self.outer_to_inner.deinit(self.circuit.allocator);
        self.inner_to_outer.deinit(self.circuit.allocator);
        self.circuit.deinit();
        allocator.destroy(self);
    }

    pub fn update(component: *Component) AllocatorError!void {
        const self: *Microchip = @fieldParentPtr("component", component);

        for (self.outer_to_inner.items, 0..) |connection_opt, idx| {
            if (connection_opt) |connection| {
                const outer = component.getInput(idx);
                const inner = &connection.component.inputs.items[connection.idx];
                inner.signal = outer;
            }
        }

        try self.circuit.tick();

        for (self.inner_to_outer.items, 0..) |connection_opt, idx| {
            if (connection_opt) |connection| {
                const outer = component.getOutput(idx);
                const inner = connection.component.getOutput(connection.idx);
                outer.* = inner.*;
            }
        }
    }

    pub fn resizeInputs(self: *Microchip, new_len: usize) !void {
        try self.component.resizeInputs(self.circuit.allocator, new_len);
        try self.outer_to_inner.resize(self.circuit.allocator, new_len);
    }

    pub fn resizeOutputs(self: *Microchip, new_len: usize) !void {
        try self.component.resizeOutputs(self.circuit.allocator, new_len);
        try self.inner_to_outer.resize(self.circuit.allocator, new_len);
    }

    pub fn connectInput(self: *Microchip, outer_idx: usize, inner: *Component, inner_idx: usize) !void {
        self.outer_to_inner.items[outer_idx] = .{
            .component = inner,
            .idx = inner_idx,
        };
    }

    pub fn connectOutput(self: *Microchip, inner: *Component, inner_idx: usize, outer_idx: usize) !void {
        self.inner_to_outer.items[outer_idx] = .{
            .component = inner,
            .idx = inner_idx,
        };
    }
};

pub const Node = struct {
    component: Component,

    pub fn init(allocator: std.mem.Allocator) !*Node {
        var self = try allocator.create(Node);
        errdefer allocator.destroy(self);
        try Component.init(&self.component, allocator, "Circuit Node", 2, 1, &update, &deinit);
        return self;
    }

    pub fn deinit(component: *Component, allocator: std.mem.Allocator) void {
        const self: *Node = @fieldParentPtr("component", component);
        allocator.destroy(self);
    }

    pub fn update(component: *Component) AllocatorError!void {
        const output = component.getOutput(0);
        const activate = component.getInput(0) orelse &Signal{ .digital = 1 };
        const trigger = component.getInput(1) orelse &Signal{};

        if (activate.digital == 0) {
            output.digital = 0;
            output.analog = 0.0;
        } else {
            output.digital = trigger.digital;
            output.analog = trigger.analog;
        }
    }
};

test "Node" {
    var trigger = Signal{ .digital = 1, .analog = 1.0 };
    var activate = Signal{};

    var node = try Node.init(std.testing.allocator);
    defer node.component.deinit(std.testing.allocator);
    node.component.inputs.items[1].signal = &trigger;

    // should always pass trigger signal if activate is disconnected
    try node.component.update();
    try std.testing.expectEqual(Signal{ .digital = 1, .analog = 1.0 }, node.component.getOutput(0).*);

    node.component.inputs.items[0].signal = &activate;

    // should not pass trigger signal if activate is 0
    activate.digital = 0;
    activate.analog = 0.0;
    try node.component.update();
    try std.testing.expectEqual(Signal{ .digital = 0, .analog = 0.0 }, node.component.getOutput(0).*);

    // should pass trigger signal if activate is 1 ...
    activate.digital = 1;
    activate.analog = 0.0;
    try node.component.update();
    try std.testing.expectEqual(Signal{ .digital = 1, .analog = 1.0 }, node.component.getOutput(0).*);

    // should not pass trigger signal if activate.digital is 0
    // even if activate.analog is not 0.0
    activate.digital = 0;
    activate.analog = 1.0;
    try node.component.update();
    try std.testing.expectEqual(Signal{ .digital = 0, .analog = 0.0 }, node.component.getOutput(0).*);

    // should preserve the trigger signal polarity and analog strength
    activate.digital = 1;
    trigger.digital = -1;
    trigger.analog = -0.69;
    try node.component.update();
    try std.testing.expectEqual(Signal{ .digital = -1, .analog = -0.69 }, node.component.getOutput(0).*);
}

pub const Battery = struct {
    component: Component,

    value: f32 = 1.0,

    pub fn init(allocator: std.mem.Allocator) !*Battery {
        var self = try allocator.create(Battery);
        errdefer allocator.destroy(self);
        self.* = .{ .component = undefined };
        try Component.init(&self.component, allocator, "Battery", 0, 1, &update, &deinit);
        return self;
    }

    pub fn deinit(component: *Component, allocator: std.mem.Allocator) void {
        const self: *Battery = @fieldParentPtr("component", component);
        allocator.destroy(self);
    }

    pub fn update(component: *Component) AllocatorError!void {
        const self: *Battery = @fieldParentPtr("component", component);
        const output = component.getOutput(0);
        output.digital = @intFromFloat(std.math.sign(self.value));
        output.analog = self.value;
    }
};

test "Battery" {
    var battery = try Battery.init(std.testing.allocator);
    defer battery.component.deinit(std.testing.allocator);

    // default in-game is 100%
    // the default should be the same here too
    try battery.component.update();
    try std.testing.expectEqual(Signal{ .digital = 1, .analog = 1.0 }, battery.component.getOutput(0).*);

    battery.value = 0.0;
    try battery.component.update();
    try std.testing.expectEqual(Signal{ .digital = 0, .analog = 0.0 }, battery.component.getOutput(0).*);

    battery.value = -1.0;
    try battery.component.update();
    try std.testing.expectEqual(Signal{ .digital = -1, .analog = -1.0 }, battery.component.getOutput(0).*);

    battery.value = 0.42;
    try battery.component.update();
    try std.testing.expectEqual(Signal{ .digital = 1, .analog = 0.42 }, battery.component.getOutput(0).*);

    battery.value = -0.69;
    try battery.component.update();
    try std.testing.expectEqual(Signal{ .digital = -1, .analog = -0.69 }, battery.component.getOutput(0).*);
}

pub const Not = struct {
    component: Component,

    invert_output: bool = true,

    pub fn init(allocator: std.mem.Allocator) !*Not {
        var self = try allocator.create(Not);
        errdefer allocator.destroy(self);
        self.* = .{ .component = undefined };
        try Component.init(&self.component, allocator, "NOT", 1, 1, &update, &deinit);
        return self;
    }

    pub fn deinit(component: *Component, allocator: std.mem.Allocator) void {
        const self: *Not = @fieldParentPtr("component", component);
        allocator.destroy(self);
    }

    pub fn update(component: *Component) AllocatorError!void {
        const self: *Not = @fieldParentPtr("component", component);
        const output = component.getOutput(0);
        const input = component.getInput(0) orelse &Signal{};
        if (self.invert_output) {
            output.digital = 1 - @as(i2, @intCast(@abs(input.digital)));
            output.analog = 1.0 - @abs(input.analog);
        } else {
            output.digital = input.digital;
            output.analog = input.analog;
        }
        output.analog = std.math.clamp(output.analog, -1.0, 1.0);
    }
};

pub const And = struct {
    component: Component,

    // if false, is in Minimum Input mode
    // if true, is in Multiply Inputs mode
    arithmetic_mode: bool = false,

    pub fn init(allocator: std.mem.Allocator) !*And {
        var self = try allocator.create(And);
        errdefer allocator.destroy(self);
        self.* = .{ .component = undefined };
        try Component.init(&self.component, allocator, "AND", 2, 1, &update, &deinit);
        return self;
    }

    pub fn deinit(component: *Component, allocator: std.mem.Allocator) void {
        const self: *And = @fieldParentPtr("component", component);
        allocator.destroy(self);
    }

    // TODO check implementation
    pub fn update(component: *Component) AllocatorError!void {
        const self: *And = @fieldParentPtr("component", component);
        const output = component.getOutput(0);
        const input0 = component.getInput(0) orelse &Signal{};
        if (self.arithmetic_mode) {
            output.digital = input0.digital;
            output.analog = input0.analog;
            for (component.inputs.items[1..]) |opt_input| {
                const input = opt_input.signal orelse &Signal{};
                output.digital = 0; // TODO
                output.analog *= input.analog;
            }
        } else {
            output.digital = input0.digital;
            output.analog = input0.analog;
            for (component.inputs.items[1..]) |opt_input| {
                const input = opt_input.signal orelse &Signal{};
                output.digital = 0; // TODO
                output.analog = switch (std.math.order(@abs(output.analog), @abs(input.analog))) {
                    .lt => output.analog,
                    .eq => @min(output.analog, input.analog), // TODO what does this *actually* do?
                    .gt => input.analog,
                };
            }
        }
        output.analog = std.math.clamp(output.analog, -1.0, 1.0);
    }
};

test "And" {
    var a = Signal{};
    var b = Signal{};

    var and1 = try And.init(std.testing.allocator);
    defer and1.component.deinit(std.testing.allocator);
    and1.component.inputs.items[0].signal = &a;
    and1.component.inputs.items[1].signal = &b;

    // TODO test digital part

    // if arithmetic_mode is false, should output the input value furthest from 0
    a.analog = 0.0;
    b.analog = 1.0;
    try and1.component.update();
    try std.testing.expectEqual(0.0, and1.component.outputs.items[0].signal.analog);

    a.analog = -0.5;
    b.analog = -0.2;
    try and1.component.update();
    try std.testing.expectEqual(-0.2, and1.component.outputs.items[0].signal.analog);

    // if arithmetic_mode is true, should output the product of all inputs
    and1.arithmetic_mode = true;

    a.analog = 0.1;
    b.analog = 0.1;
    try and1.component.update();
    try std.testing.expectApproxEqAbs(0.01, and1.component.outputs.items[0].signal.analog, 0.00001);
}

pub const Or = struct {
    component: Component,

    // if false, is in Maximum Input mode
    // if true, is in Add Inputs mode
    arithmetic_mode: bool = false,

    pub fn init(allocator: std.mem.Allocator) !*Or {
        var self = try allocator.create(Or);
        errdefer allocator.destroy(self);
        self.* = .{ .component = undefined };
        try Component.init(&self.component, allocator, "OR", 2, 1, &update, &deinit);
        return self;
    }

    pub fn deinit(component: *Component, allocator: std.mem.Allocator) void {
        const self: *Or = @fieldParentPtr("component", component);
        allocator.destroy(self);
    }

    // TODO check implementation
    pub fn update(component: *Component) AllocatorError!void {
        const self: *Or = @fieldParentPtr("component", component);
        const output = component.getOutput(0);
        const input0 = component.getInput(0) orelse &Signal{};
        if (self.arithmetic_mode) {
            output.digital = input0.digital;
            output.analog = input0.analog;
            for (component.inputs.items[1..]) |opt_input| {
                const input = opt_input.signal orelse &Signal{};
                output.digital = 0; // TODO
                output.analog += input.analog;
            }
        } else {
            output.digital = input0.digital;
            output.analog = input0.analog;
            for (component.inputs.items[1..]) |opt_input| {
                const input = opt_input.signal orelse &Signal{};
                output.digital = 0; // TODO
                output.analog = switch (std.math.order(@abs(output.analog), @abs(input.analog))) {
                    .lt => input.analog,
                    .eq => @max(output.analog, input.analog), // TODO what does this *actually* do?
                    .gt => output.analog,
                };
            }
        }
        output.analog = std.math.clamp(output.analog, -1.0, 1.0);
    }
};

test "Or" {
    var a = Signal{ .analog = 0.0 };
    var b = Signal{ .analog = 1.0 };

    var or1 = try Or.init(std.testing.allocator);
    defer or1.component.deinit(std.testing.allocator);
    or1.component.inputs.items[0].signal = &a;
    or1.component.inputs.items[1].signal = &b;

    // TODO test digital part

    try or1.component.update();
    try std.testing.expectEqual(1.0, or1.component.getOutput(0).analog);

    a.analog = -0.5;
    b.analog = -0.2;
    try or1.component.update();
    try std.testing.expectEqual(-0.5, or1.component.getOutput(0).analog);

    // TODO arithmetic mode
}

pub const Selector = struct {
    component: Component,

    current_output: usize = 0,
    invert_output: bool = false,

    pub fn init(allocator: std.mem.Allocator) !*Selector {
        var self = try allocator.create(Selector);
        errdefer allocator.destroy(self);
        self.* = .{ .component = undefined };
        try Component.init(&self.component, allocator, "Selector", 5, 4, &update, &deinit);
        return self;
    }

    pub fn deinit(component: *Component, allocator: std.mem.Allocator) void {
        const self: *Selector = @fieldParentPtr("component", component);
        allocator.destroy(self);
    }

    // TODO check implementation
    // TODO broken, needs to be completely rewritten
    pub fn update(component: *Component) AllocatorError!void {
        const self: *Selector = @fieldParentPtr("component", component);

        const cycle = component.getInput(0) orelse &Signal{};

        std.debug.print("len {d}\n", .{component.inputs.items.len});

        if (cycle.digital > 0) {
            self.increment();
        } else if (cycle.digital < 0) {
            self.decrement();
        } else {
            var last_idx: ?usize = null;
            for (component.inputs.items[1..], 0..) |input, idx| {
                if (input.signal) |signal| {
                    if (signal.digital == 1) last_idx = idx;
                }
            }
            if (last_idx) |lidx| {
                for (component.outputs.items[0..], 0..) |*output, idx| {
                    const flip = if (self.invert_output)
                        idx != lidx
                    else
                        idx == lidx;

                    output.signal = if (flip)
                        .{ .digital = 1, .analog = 1.0 }
                    else
                        .{ .digital = 0, .analog = 0.0 };
                }
            } else {
                for (component.outputs.items[0..], 0..) |*output, idx| {
                    const flip = if (self.invert_output)
                        idx != self.current_output
                    else
                        idx == self.current_output;

                    std.debug.print("{d} flip {}\n", .{ idx, flip });

                    output.signal = if (flip)
                        .{ .digital = 1, .analog = 1.0 }
                    else
                        .{ .digital = 0, .analog = 0.0 };
                }
            }
        }

        std.debug.print("final values", .{});
        for (component.outputs.items) |output| {
            std.debug.print(" {}", .{output.signal});
        }
        std.debug.print("\n", .{});
    }

    pub fn resize(self: *Selector, allocator: std.mem.Allocator, new_len: usize) !void {
        try self.component.resizeInputs(allocator, new_len + 1);
        try self.component.resizeOutputs(allocator, new_len);
    }

    fn increment(self: *Selector) void {
        const len = self.component.outputs.items.len;
        if (self.current_output >= len)
            self.current_output = 0
        else
            self.current_output += 1;
    }

    fn decrement(self: *Selector) void {
        const len = self.component.outputs.items.len;
        if (self.current_output >= 0)
            self.current_output = len
        else
            self.current_output -= 1;
    }
};

test "Selector" {
    var a = Signal{};
    var b = Signal{};
    var c = Signal{};

    var selector = try Selector.init(std.testing.allocator);
    defer selector.component.deinit(std.testing.allocator);
    try selector.resize(std.testing.allocator, 3);

    // default state, output 0 should be active
    try selector.component.update();
    try std.testing.expectEqual(Signal{ .digital = 1, .analog = 1.0 }, selector.component.getOutput(0).*);
    try std.testing.expectEqual(Signal{ .digital = 0, .analog = 0.0 }, selector.component.getOutput(1).*);
    try std.testing.expectEqual(Signal{ .digital = 0, .analog = 0.0 }, selector.component.getOutput(2).*);

    // should remember last active output
    selector.component.inputs.items[1].signal = &a;
    selector.component.inputs.items[2].signal = &b;
    selector.component.inputs.items[3].signal = &c;

    a.digital = 0;
    b.digital = 1;
    c.digital = 0;
    try selector.component.update();
    try std.testing.expectEqual(Signal{ .digital = 0, .analog = 0.0 }, selector.component.getOutput(0).*);
    try std.testing.expectEqual(Signal{ .digital = 1, .analog = 1.0 }, selector.component.getOutput(1).*);
    try std.testing.expectEqual(Signal{ .digital = 0, .analog = 0.0 }, selector.component.getOutput(2).*);

    a.digital = 0;
    b.digital = 0;
    c.digital = 0;
    try selector.component.update();
    try std.testing.expectEqual(Signal{ .digital = 0, .analog = 0.0 }, selector.component.getOutput(0).*);
    try std.testing.expectEqual(Signal{ .digital = 1, .analog = 1.0 }, selector.component.getOutput(1).*);
    try std.testing.expectEqual(Signal{ .digital = 0, .analog = 0.0 }, selector.component.getOutput(2).*);

    selector.invert_output = true;
    try selector.component.update();
    try std.testing.expectEqual(Signal{ .digital = 1, .analog = 1.0 }, selector.component.getOutput(0).*);
    try std.testing.expectEqual(Signal{ .digital = 0, .analog = 0.0 }, selector.component.getOutput(1).*);
    try std.testing.expectEqual(Signal{ .digital = 1, .analog = 1.0 }, selector.component.getOutput(2).*);

    // TODO cycle input
}

const AllocatorError = std.mem.Allocator.Error;
